@@ -1,7 +1,7 @@
 
 /* xorriso - creates, loads, manipulates and burns ISO 9660 filesystem images.
 
-   Copyright 2007-2013 Thomas Schmitt, <scdbackup@gmx.net>
+   Copyright 2007-2016 Thomas Schmitt, <scdbackup@gmx.net>
 
    Provided under GPL version 2 or later.
 
@@ -96,7 +96,7 @@ int Xorriso_opt_args(struct XorrisO *xorriso, char *cmd,
  else
    do_expand= (xorriso->do_iso_rr_pattern==1 && !(flag&4)) || (flag & 512);
  if(flag&256) {
-   if(*optv<argv || *optv>=argv+argc)
+   if(*optv < argv || (*optv >= argv + argc && argc > 0))
      Sfile_destroy_argv(optc, optv, 0);
    return(1);
  }
@@ -197,14 +197,24 @@ ex:;
 }
 
 
+/* @param flag bit0= get eternal problem status
+*/
 int Xorriso_get_problem_status(struct XorrisO *xorriso, char severity[80],
                                int flag)
 {
- strcpy(severity, xorriso->problem_status_text);
- return(xorriso->problem_status);
+ if(flag & 1) {
+   strcpy(severity, xorriso->eternal_problem_status_text);
+   return(xorriso->eternal_problem_status);
+ } else {
+   strcpy(severity, xorriso->problem_status_text);
+   return(xorriso->problem_status);
+ }
 }
 
 
+/* @param flag bit0= set eternal problem status to severity,
+                     and set problem status to ALL
+*/
 int Xorriso_set_problem_status(struct XorrisO *xorriso, char *severity, 
                                int flag)
 {
@@ -216,7 +226,7 @@ int Xorriso_set_problem_status(struct XorrisO *xorriso, char *severity,
  static int complaints= 0, complaint_limit= 5;
 #endif
 
- if(severity[0])
+ if(severity[0] && strlen(severity) < sizeof(xorriso->problem_status_text))
    sev_text= severity;
  ret= Xorriso__text_to_sev(sev_text, &sev, 0);
  if(ret<=0)
@@ -237,9 +247,15 @@ int Xorriso_set_problem_status(struct XorrisO *xorriso, char *severity,
 
 #endif /* Xorriso_fetch_with_msg_queueS */
 
- xorriso->problem_status= sev;
- strcpy(xorriso->problem_status_text, sev_text);
- if(sev > xorriso->eternal_problem_status) {
+ if(flag & 1) {
+   strcpy(xorriso->problem_status_text, "ALL");
+   Xorriso__text_to_sev(xorriso->problem_status_text,
+                        &(xorriso->problem_status), 0);
+ } else {
+   xorriso->problem_status= sev;
+   strcpy(xorriso->problem_status_text, sev_text);
+ }
+ if(sev > xorriso->eternal_problem_status || (flag & 1)) {
    xorriso->eternal_problem_status= sev;
    strcpy(xorriso->eternal_problem_status_text, sev_text);
  }
@@ -429,12 +445,12 @@ int Xorriso_decode_load_adr(struct XorrisO *xorriso, char *cmd,
 }
 
  
-int Xorriso_check_name_len(struct XorrisO *xorriso, char *name, int size,
-                           char *cmd, int flag)
+int Xorriso_check_thing_len(struct XorrisO *xorriso, char *name, int size,
+                            char *cmd, char *thing, int flag)
 {
  if((int) strlen(name) >= size) {
    sprintf(xorriso->info_text,
-           "Name too long with option %s (%d > %d)", cmd,
+           "%s too long with option %s (%d > %d)", thing, cmd,
            (int) strlen(name), size - 1);
    Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "SORRY", 0);
    return(0);
@@ -443,6 +459,12 @@ int Xorriso_check_name_len(struct XorrisO *xorriso, char *name, int size,
 }
  
 
+int Xorriso_check_name_len(struct XorrisO *xorriso, char *name, int size,
+                           char *cmd, int flag)
+{
+ return Xorriso_check_thing_len(xorriso, name, size, cmd, "Name", flag);
+}
+
 
 /* @return <0 error , >=0 number of skipped dashes
 */
@@ -506,17 +528,20 @@ int Xorriso_count_args(struct XorrisO *xorriso, int argc, char **argv,
     "close","close_damaged",
     "commit_eject","compliance","copyright_file",
     "dev","dialog","disk_dev_ino","disk_pattern","displacement",
-    "dummy","dvd_obs","early_stdio_test", "eject",
-    "iso_rr_pattern","follow","format","fs","gid","grow_blindly","hardlinks",
+    "dummy","dvd_obs","early_stdio_test","ecma119_map","eject",
+    "iso_rr_pattern","file_name_limit","follow","format","fs",
+    "gid","grow_blindly","hardlinks",
     "hfsplus","history","indev","in_charset","joliet",
     "list_delimiter","list_extras","list_profiles","local_charset",
-    "mark","md5","mount_opts","not_leaf","not_list","not_mgt",
+    "mark","md5","mount_opts","modesty_on_drive",
+    "not_leaf","not_list","not_mgt",
     "options_from_file","osirrox","outdev","out_charset","overwrite",
     "pacifier","padding","path_list","pathspecs","pkt_output",
     "preparer_id","print","print_info","print_mark","prompt",
     "prog","prog_help","publisher","quoted_not_list","quoted_path_list",
-    "reassure","report_about","rockridge",
-    "rom_toc_scan","rr_reloc_dir","scsi_log",
+    "read_fs","read_speed","reassure","report_about",
+    "report_el_torito","report_system_area","rockridge",
+    "rom_toc_scan","rr_reloc_dir","scsi_dev_family","scsi_log",
     "session_log","sh_style_result","signal_handling","sleep",
     "speed","split_size","status","status_history_max",
     "stdio_sync","stream_recording","system_id","temp_mem_limit","toc_of",
@@ -547,7 +572,7 @@ int Xorriso_count_args(struct XorrisO *xorriso, int argc, char **argv,
     "check_md5","check_md5_r","check_media","check_media_defaults",
     "chgrp","chgrpi","chgrp_r","chgrp_ri","chmod","chmodi",
     "chmod_r","chmod_ri","chown","chowni","chown_r","chown_ri",
-    "compare_l","cp_clone","cp_rax","cp_rx","cpr","cpri", "cpax","cpx",
+    "compare_l","concat","cp_clone","cp_rax","cp_rx","cpr","cpri","cpax","cpx",
     "du","dui","dus","dusi","dux","dusx","external_filter","extract_l",
     "file_size_limit","find","findi","finds","findx",
     "getfacl","getfacli","getfacl_r","getfacl_ri",
@@ -649,10 +674,13 @@ int Xorriso_cmd_sorting_rank(struct XorrisO *xorriso,
    "mount_opts", "mount_cmd", "session_string",
 
    "* Influencing the behavior of image loading:",
-   "load", "displacement", "drive_class", "assert_volid", "in_charset",
+   "read_speed", "load", "displacement", "read_fs",
+   "drive_class", "assert_volid", "in_charset",
    "auto_charset", "hardlinks", "acl", "xattr", "md5", "for_backup",
+   "ecma119_map",
    "disk_dev_ino", "rom_toc_scan", "calm_drive", "ban_stdio_write",
    "early_stdio_test", "data_cache_size",
+   "scsi_dev_family",
 
    "* Character sets:",
    "charset", "local_charset",
@@ -662,10 +690,11 @@ int Xorriso_cmd_sorting_rank(struct XorrisO *xorriso,
 
    "* Drive and media related inquiry actions (2):",
    "list_profiles", "list_formats", "list_speeds",
-   "toc", "toc_of", "pvd_info",
+   "toc", "toc_of", "pvd_info", "report_system_area", "report_el_torito",
 
    "* Settings for file insertion:",
-   "file_size_limit", "not_mgt", "not_paths", "not_leaf", "not_list",
+   "file_name_limit", "file_size_limit",
+   "not_mgt", "not_paths", "not_leaf", "not_list",
    "quoted_not_list", "follow", "pathspecs", "overwrite", "split_size",
 
    "* Navigation in ISO image and disk filesystem (1):",
@@ -700,7 +729,7 @@ int Xorriso_cmd_sorting_rank(struct XorrisO *xorriso,
 
    "* osirrox ISO-to-disk restore options:",
    "osirrox", "extract", "extract_single", "extract_l", "extract_cut",
-   "cpx", "cpax", "cp_rx", "cp_rax", "paste_in",
+   "cpx", "cpax", "cp_rx", "cp_rax", "paste_in", "concat",
    "mount",
 
    "* Settings for result writing:",
@@ -710,6 +739,7 @@ int Xorriso_cmd_sorting_rank(struct XorrisO *xorriso,
    "abstract_file", "biblio_file", "preparer_id", "application_use",
    "out_charset", "read_mkisofsrc",
    "uid", "gid", "zisofs", "speed", "stream_recording", "dvd_obs",
+   "modesty_on_drive",
    "stdio_sync", "dummy", "fs", "close", "padding", "write_type",
    "grow_blindly", "pacifier", "scdbackup_tag",
 
@@ -1147,6 +1177,9 @@ next_command:;
    (*idx)++;
    Xorriso_option_compliance(xorriso, arg1, 0);
 
+ } else if(strcmp(cmd,"concat") == 0) {
+   ret= Xorriso_option_concat(xorriso, argc, argv, idx, 0);
+
  } else if(strcmp(cmd,"copyright_file")==0) { 
    (*idx)++;
    Xorriso_option_copyright_file(xorriso, arg1, 0);
@@ -1229,6 +1262,10 @@ next_command:;
    (*idx)++;
    ret= Xorriso_option_early_stdio_test(xorriso, arg1, 0);
 
+ } else if(strcmp(cmd,"ecma119_map")==0) {
+   (*idx)++;
+   ret= Xorriso_option_ecma119_map(xorriso, arg1, 0);
+
  } else if(strcmp(cmd,"eject")==0) {
    (*idx)++;
    ret= Xorriso_option_eject(xorriso, arg1, 0);
@@ -1274,6 +1311,10 @@ next_command:;
    (*idx)+= 2;
    ret= Xorriso_option_extract(xorriso, arg1, arg2, 32);
 
+ } else if(strcmp(cmd,"file_name_limit")==0) {
+   (*idx)++;
+   ret= Xorriso_option_file_name_limit(xorriso, arg1, 0);
+
  } else if(strcmp(cmd,"file_size_limit")==0) {
    ret= Xorriso_option_file_size_limit(xorriso, argc, argv, idx, 0);
 
@@ -1435,6 +1476,13 @@ next_command:;
    (*idx)++;
    ret= Xorriso_option_md5(xorriso, arg1, 0);
 
+ } else if(strcmp(cmd,"mkdir")==0 || strcmp(cmd,"mkdiri")==0) {
+   ret= Xorriso_option_mkdiri(xorriso, argc, argv, idx, 0);
+
+ } else if(strcmp(cmd, "modesty_on_drive")==0) {
+   (*idx)++;
+   ret= Xorriso_option_modesty_on_drive(xorriso, arg1, 0);
+
  } else if(strcmp(cmd, "mount") == 0 || strcmp(cmd, "mount_cmd") == 0) {
    (*idx)+= 4;
    if((*idx)>argc) {
@@ -1463,9 +1511,6 @@ next_command:;
  } else if(strcmp(cmd,"mv")==0 || strcmp(cmd,"mvi")==0) {
    ret= Xorriso_option_mvi(xorriso, argc, argv, idx, 0);
 
- } else if(strcmp(cmd,"mkdir")==0 || strcmp(cmd,"mkdiri")==0) {
-   ret= Xorriso_option_mkdiri(xorriso, argc, argv, idx, 0);
-
  } else if(strcmp(cmd,"named_pipe_loop")==0) {
    if((*idx) + 3 < argc)
      arg4= argv[(*idx) + 3];
@@ -1608,9 +1653,17 @@ next_command:;
    (*idx)++;
    ret= Xorriso_option_path_list(xorriso, arg1, 1);
 
+ } else if(strcmp(cmd,"read_fs")==0) {
+   (*idx)++;
+   ret= Xorriso_option_read_fs(xorriso, arg1, 0);
+
  } else if(strcmp(cmd,"read_mkisofsrc")==0) {
    ret= Xorriso_option_read_mkisofsrc(xorriso, 0);
 
+ } else if(strcmp(cmd,"read_speed")==0) {
+   (*idx)++;
+   ret= Xorriso_option_speed(xorriso, arg1, 1);
+
  } else if(strcmp(cmd,"reassure")==0) {
    (*idx)++;
    ret= Xorriso_option_reassure(xorriso, arg1, 0);
@@ -1619,6 +1672,14 @@ next_command:;
    (*idx)++;
    ret= Xorriso_option_report_about(xorriso, arg1, 0);
 
+ } else if(strcmp(cmd,"report_el_torito")==0) {
+   (*idx)++;
+   ret= Xorriso_option_report_el_torito(xorriso, arg1, 0);
+
+ } else if(strcmp(cmd,"report_system_area")==0) {
+   (*idx)++;
+   ret= Xorriso_option_report_system_area(xorriso, arg1, 0);
+
  } else if(strcmp(cmd,"return_with")==0) {
    (*idx)+= 2;
    num2= 0;
@@ -1661,6 +1722,10 @@ next_command:;
    (*idx)+= 2;
    ret= Xorriso_option_scdbackup_tag(xorriso, arg1, arg2, 0);
  
+ } else if(strcmp(cmd, "scsi_dev_family") == 0) {
+   (*idx)++;
+   ret= Xorriso_option_scsi_dev_family(xorriso, arg1, 0);
+
  } else if(strcmp(cmd, "scsi_log") == 0) {
    (*idx)++;
    ret= Xorriso_option_scsi_log(xorriso, arg1, 0);
@@ -2012,9 +2077,8 @@ int Xorriso_execute_option(struct XorrisO *xorriso, char *line, int flag)
  char **argv= NULL;
  double tdiff;
  struct timeval tv;
- struct timezone tz;
 
- gettimeofday(&tv,&tz);
+ gettimeofday(&tv, NULL);
  Xorriso_reset_counters(xorriso,0);
  xorriso->idle_time= 0.0;
  tdiff= tv.tv_sec+(1.e-6*(double) tv.tv_usec);
@@ -2031,7 +2095,7 @@ int Xorriso_execute_option(struct XorrisO *xorriso, char *line, int flag)
  ret= Xorriso_interpreter(xorriso, argc, argv, &idx, flag&0xffff);
  if(ret<0)
    goto ex;
- gettimeofday(&tv,&tz);
+ gettimeofday(&tv, NULL);
  tdiff= tv.tv_sec+(1.e-6*(double) tv.tv_usec)-tdiff-xorriso->idle_time;
  if(tdiff<0.001)
    tdiff= 0.001;
@@ -2421,6 +2485,8 @@ int Xorriso_read_as_mkisofsrc(struct XorrisO *xorriso, char *path, int flag)
      /* >>> ??? complain ? abort reading ? */;
 
    }
+   if(ret <= 0)
+     goto ex;
  }
  xorriso->mkisofsrc_done= 1;
  ret= 1;
@@ -2543,7 +2609,7 @@ int Xorriso_program_arg_bsl(struct XorrisO *xorriso, int argc, char ***argv,
  Xorriso_alloc_meM(cmd_data, char, cmd_data_size);
  new_argv= (char **) Smem_malloC(argc * sizeof(char *));
  if(new_argv == NULL)
-   return(-1); 
+   {ret= -1; goto ex;}
  for(i= 0; i < argc; i++) {
    new_argv[i]= strdup((*argv)[i]);
    if(new_argv[i] == NULL)
@@ -2654,12 +2720,21 @@ int Xorriso_convert_datestring(struct XorrisO *xorriso, char *cmd,
 {
  int ret;
 
+ *t_type= 0;
  if(strcmp(time_type, "a")==0)
    (*t_type)|= 1;
+ else if(strcmp(time_type, "a-c")==0)
+   (*t_type)|= 1 | 256;
  else if(strcmp(time_type, "m")==0)
    (*t_type)|= 4;
+ else if(strcmp(time_type, "m-c")==0)
+   (*t_type)|= 4 | 256;
  else if(strcmp(time_type, "b")==0)
    (*t_type)|= 5;
+ else if(strcmp(time_type, "b-c")==0)
+   (*t_type)|= 5 | 256;
+ else if(strcmp(time_type, "c")==0)
+   (*t_type)|= 2 | 256;
  else {
    sprintf(xorriso->info_text, "%s: Unrecognized type '%s'", cmd, time_type);
    if(!(flag & 1))
@@ -2705,16 +2780,67 @@ int Xorriso_check_temp_mem_limit(struct XorrisO *xorriso, off_t mem, int flag)
 }
 
 
+int Xorriso_wait_child_end(struct XorrisO *xorriso, pid_t child_pid,
+                           int *status, int flag)
+{
+ int ret;
+
+ do {
+   /* try to read and print the reply */;
+   ret= waitpid(child_pid, status, WNOHANG);
+   if(ret == -1) {
+     if(errno != EINTR)
+       return(0);
+   } else if(ret == 0) {
+ continue;
+   } else {
+ break;
+   }
+ } while(1);
+
+ /* >>> interpret *status */;
+
+ return(1);
+}
+
+
+int Xorriso_make_argv_with_null(struct XorrisO *xorriso,
+                                int in_argc, char **in_argv,
+                                int *argc, char ***argv, int flag)
+{
+ int i, ret= 0;
+
+ *argv= NULL;
+ Xorriso_alloc_meM(*argv, char *, in_argc + 1);
+ for(i= 0; i < in_argc; i++) {
+   Xorriso_alloc_meM((*argv)[i], char, strlen(in_argv[i]) + 1);
+   strcpy((*argv)[i], in_argv[i]);
+   *argc= i + 1;
+ }
+ (*argv)[in_argc]= NULL;
+ ret= 1;
+ex:;
+ if(ret <= 0)
+   Sfile_destroy_argv(argc, argv, 0);
+ return(ret);
+}
+
+
 /*
   @param flag bit0= use env_path to find the desired program
+              bit1= use in_argv rather than parsing cmd to words
+              bit2= -reserved-
+              bit3= demand absolute cmd path
 return:
  <=0 : error
  1   : done
 */
-int Xorriso_execv(struct XorrisO *xorriso, char *cmd, char *env_path,
+int Xorriso_execv(struct XorrisO *xorriso, char *cmd,
+                  int in_argc, char **in_argv, char *env_path,
+                  int *stdin_pipe, int *stdout_pipe, pid_t *forked_pid,
                   int *status, int flag)
 {
- int ret, argc= 0;
+ int ret, argc= 0, has_slash;
  char **argv= NULL, *pathlist= NULL, *cpt, *npt, *prog= NULL;
  pid_t child_pid;
  struct stat stbuf;
@@ -2723,14 +2849,26 @@ int Xorriso_execv(struct XorrisO *xorriso, char *cmd, char *env_path,
 
  wait3(NULL,WNOHANG,NULL); /* just to remove any old dead child */
 
- ret= Sfile_make_argv("", cmd, &argc, &argv, 1|4|128);
+ if(flag & 2) {
+   ret= Xorriso_make_argv_with_null(xorriso, in_argc, in_argv,
+                                    &argc, &argv, 0);
+ } else {
+   ret= Sfile_make_argv("", cmd, &argc, &argv, 1|4|128);
+ }
  if(ret <= 0)
    goto ex;
  if(argc < 1)
    {ret= 0; goto ex;}
 
  strcpy(prog, argv[0]);
- if((flag & 1) && strchr(argv[0], '/') == NULL) {
+ has_slash= (strchr(argv[0], '/') != NULL);
+ if((flag & 8) && !has_slash) {
+   sprintf(xorriso->info_text, "External program path contains no '/': ");
+   Text_shellsafe(argv[0], xorriso->info_text, 1);
+   Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "FAILURE", 0);
+   ret= 0; goto ex;
+ }
+ if((flag & 1) && !has_slash) {
      if(env_path == NULL)
        env_path= "/bin:/sbin";
      else if(env_path[0] == 0)
@@ -2759,7 +2897,7 @@ int Xorriso_execv(struct XorrisO *xorriso, char *cmd, char *env_path,
 
  child_pid= fork();
  if(child_pid==-1) 
-   return(-1);
+   {ret= -1; goto ex;}
 
  if(child_pid==0) {
                      /* this is the child process */
@@ -2767,9 +2905,21 @@ int Xorriso_execv(struct XorrisO *xorriso, char *cmd, char *env_path,
    sprintf(xorriso->info_text, "Executing external program ");
    Text_shellsafe(prog, xorriso->info_text, 1);
    Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "DEBUG", 0);
-
    Xorriso_destroy(&xorriso, 0); /* reduce memory foot print */
 
+   if(stdin_pipe != NULL) {
+     close(0);
+     if(dup2(stdin_pipe[0], 0) == -1)
+            { perror("dup2(,0)"); exit(1); }
+     close(stdin_pipe[1]); /* unused */
+   }
+   if(stdout_pipe != NULL) {
+     close(1);
+     if(dup2(stdout_pipe[1], 1) == -1)
+            { perror("dup2(,1)"); exit(1); }
+     close(stdout_pipe[0]); /* unused */
+   }
+
    execv(prog, argv); /* should never come back */
    fprintf(stderr,"--- execution of shell command failed:\n");
    fprintf(stderr,"    %s\n",cmd);
@@ -2778,33 +2928,22 @@ int Xorriso_execv(struct XorrisO *xorriso, char *cmd, char *env_path,
 
 
          /* this is the original process waiting for child to exit */
- do {
-   /* try to read and print the reply */;
-   ret= waitpid(child_pid,status,WNOHANG);
-   if(ret==-1) {
-     if(errno!=EINTR)
-       ret= 0; goto ex;
-   } else if(ret==0) {
 
-#ifdef NIX
+ if(stdin_pipe != NULL)
+   close(stdin_pipe[0]); /* unused */
+ if(stdout_pipe != NULL)
+   close(stdout_pipe[1]); /* unused */
+ if(stdin_pipe != NULL || stdout_pipe != NULL) {
+   /* Pipes need to be fed by the caller who later calls Xorriso_wait_child_end
+      with *forked_pid as child_pid.
+   */
+   *forked_pid= child_pid;
+   {ret= 1; goto ex;}
+ }
 
-     /* >>> An interrupt key would be nice. */
-
-     if((flag&4)) {
-       ret= Asqueue_event_is_pending(agent->queue,0,0);
-       if(ret>0) {
-         Asagent_stderr(agent,"--- shell command interrupted",1);
-         kill(child_pid,SIGTERM);
-         ret= 2; goto ex;
-       }
-     }
-#endif /* NIX */
-
- continue;
-   } else {
- break;
-   }
- } while(1);
+ ret= Xorriso_wait_child_end(xorriso, child_pid, status, 0);
+ if(ret <= 0)
+   goto ex;
  ret= 1;
 ex:
  Sfile_make_argv("", "", &argc, &argv, 2);
@@ -2814,6 +2953,43 @@ ex:
 }
 
 
+/* @param flag bit0= use env_path to find the desired program
+               bit1= use in_argv rather than parsing cmd to words
+               bit2= "r" rather than "w"
+               bit3= demand absolute cmd path
+               bit4= override any restriction on external filters
+*/
+int Xorriso_pipe_open(struct XorrisO *xorriso, char *purpose, char *cmd,
+                      int in_argc, char **in_argv, char *env_path,
+                      int *fd, pid_t *forked_pid, int flag)
+{
+ int fp_pipe[2], *stdin_pipe= NULL, *stdout_pipe= NULL, status, ret;
+
+ *fd= -1;
+
+ if(!(flag & 16)) {
+   ret= Xorriso_external_filter_banned(xorriso, purpose, 0);
+   if(ret)
+     return(0);
+ }
+ if(pipe(fp_pipe) != 0) {
+   sprintf(xorriso->info_text, "Cannot create pipe(2) object");
+   Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, errno, "FATAL", 0);
+   return(0);
+ }
+ if(flag & 4) {
+   stdout_pipe= fp_pipe;
+   *fd= fp_pipe[0];
+ } else {
+   stdin_pipe= fp_pipe;
+   *fd= fp_pipe[1];
+ }
+ ret= Xorriso_execv(xorriso, cmd, in_argc, in_argv, env_path,
+                    stdin_pipe, stdout_pipe, forked_pid, &status, flag & 11);
+ return(ret);
+}
+
+
 /* @param flag bit0= path is a command parameter
 */
 int Xorriso_path_is_excluded(struct XorrisO *xorriso, char *path, int flag)
